#!/usr/bin/python
import serial
from time import sleep, time

time0=time()

def xclock():
    return "%8.3f" % (time() - time0)

def ParseEvent(frame, verbose=0):
    address = frame[0]
    length = frame[1]
    command = frame[2]
    params =  frame[3:length-2]
    if command != 123:
        return None, None
    if params[6] & 0x80:
        if verbose: print(xclock(),"EVENT: %d.%d%d %d:%d:%d Button pressed" % (params[0],params[1],params[2], params[3],params[4],params[5]))
        return None, 1
    else:
        code = "%02x%02x%02x%02x%02x" % (params[7],params[8],params[9], params[10],params[11])
        if code [-2:] == '00' : code = code[:-2] 
        if verbose: print(xclock(),"EVENT: %d.%d%d %d:%d:%d Card %s" % (
                        params[0],params[1],params[2], params[3],params[4],params[5],
                        code))
        return code, None


def PrintFrame(frame):
    address = frame[0]
    length = frame[1]
    command = frame[2]
    params =  frame[3:length-2]
    if command == 19:
        print("CARD: ", [ "%02x" % x for x in params ])
    elif command == 123:
        if params[6] & 0x80:
            print(xclock(),"EVENT: %d.%d%d %d:%d:%d Button pressed" % (params[0],params[1],params[2], params[3],params[4],params[5]))
        else:
            print(xclock(),"EVENT: %d.%d%d %d:%d:%d Card %02x%02x%02x%02x%02x" % (
                        params[0],params[1],params[2], params[3],params[4],params[5],
                        params[7],params[8],params[9], params[10],params[11]))
    else:
        print(xclock(),"CMD:", command, params)


def LicxCRC2(bytes):

    result = 0;
    for B in bytes: 
        C=((result >> 8) ^ B) << 8
        for i in range(8):
            if (C & 0x8000): C = (C << 1) ^ 0x1021
            else: C = C << 1
        result = C ^ (result << 8)
    return result & 0xFFFF



class NetronixBus:

    BROADCAST_ADDR=255

    def __init__(self, port, baudrate=9600, timeout=0.1, verbose=0):
        self.port = serial.Serial(port=port, baudrate=baudrate, timeout=timeout)
        self.verbose = verbose


    def ClearBuffer(self):
        timeout = self.port.timeout
        self.port.timeout = 0.01
        buf = self.port.read(100)
        if (len(buf) > 0) and (self.verbose > 1):
            print(xclock(),"CLEAR:", [ "%02x" % ord(x) for x in buf ])
        while len(buf) == 100:
            buf = self.port.read(100)
            if (len(buf) > 0) and (self.verbose > 1):
                print(xclock(),"CLEAR:", [ "%02x" % ord(x) for x in buf ])
        self.port.timeout = timeout



    def ReadFrame(self):
        header = self.port.read(2)
        if len(header) != 2: 
            if self.verbose > 0:
                print(xclock(),[ "%02x" % ord(x) for x in header ])
            self.ClearBuffer()
            return None
        frame = [ ord(x) for x in header ]
        length = frame[1]
        rest = self.port.read(length-2)
        if (length-2) != len(rest):
            if self.verbose > 0:
                print("Invalid frame length")
                print(xclock(),[ "%02x" % x for x in frame ])
            self.ClearBuffer()
            return None
        frame +=  [ ord(x) for x in rest ]
        crc = (frame[length-2] << 8) + frame[length-1]
        if self.verbose > 1 : PrintFrame(frame)
        if crc != LicxCRC2(frame[:length-2]):
            if self.verbose > 0:
                print("Invalid CRC")
            return None
        return frame


    def do_command(self, address, command, params):
        self.ClearBuffer()
        frame = [address, len(params) + 5, command] + params
        crc = LicxCRC2(frame)
        frame = frame + [ crc >> 8, crc & 0xff ]
        if self.verbose > 1 : 
            print(xclock(),"SEND:", command)
        self.port.write(frame)
        recvframe =self. ReadFrame()
        if recvframe and (recvframe[2] == command+1):  # check return opcode
            return recvframe
        return None


    def ScanBus(self):
        aset = set()
        timeout = self.port.timeout
        try:
            while self.ReadFrame(): pass  # clear recv buffer

            self.port.timeout = 1
            frame = self.do_command(self.BROADCAST_ADDR, 0x78, [])
            while frame:
                if (frame[2] != 0x79): break
                address = frame[0]
                aset.add(address)
                frame = self.ReadFrame()

        finally:
            self.port.timeout = timeout

        devices = [ NetronixDevice(self,address=address) for address in aset ]
        return devices


    def GetEventStatus(self, devices):

        maxaddr = max([d.address for d in devices])
        Frames = [None,] * (maxaddr+1)
        frame = self.do_command(self.BROADCAST_ADDR, 0x78, [])
        if not frame: return

        while True:
            address = frame[0]
            if address <= maxaddr: Frames[address] = frame
            if address >= maxaddr: break
            frame = self.ReadFrame()
            if (frame is None) or (frame[2] != 0x79): break

        sleep(0.02)
        for device in devices:
            if Frames[device.address] != None: 
                device.Event(Frames[device.address])




class NetronixDevice:

    def __init__(self, bus, address):
        self.bus = bus
        self.address = address
        self.lastEvent = None
        if not hasattr(self,'Worker'):
            self.Worker = None


    def Event(self, frame):
        pointer = frame[5] + (frame[6] << 8)
        capacity = frame[3] + (frame[4] << 8)
        counter = frame[7] | (frame[8] << 8) | (frame[9] << 16) | (frame[10] << 24)
        if counter == 0xffffffff: 
            counter = 0 
        else:
            counter += 1
        if not(self.lastEvent is None):
            #print "EVENT", self.address, capacity, pointer, counter, self.lastEvent, self.lastCounter
            if counter > self.lastCounter:
                while self.lastCounter < counter: 
                    #print "GET EVENT", self.address, self.lastEvent
                    frame =  self.GetEvent(self.lastEvent)
                    if frame is None: 
                        frame =  self.GetEvent(self.lastEvent)
                    if frame:
                        if self.Worker:
                            self.Worker(self.address, ParseEvent(frame))
                        else:
                            PrintFrame(frame)
                    self.lastEvent = (self.lastEvent + 1) % capacity 
                    self.lastCounter += 1

        self.lastEvent = pointer
        self.lastCounter = counter
    

    def Login(self):
        self.bus.do_command(self.address, 0xb2, [ord('1'), ord('2'), ord('3'), ord('4'), 0])


    def SetAutowrite(self, enable=False):
        self.bus.do_command(self.address, 0x74, [(1 if enable else 0)])


    def SetIOConfig(self, io, mode, time, time0, time1):
        # io 2=green, 3=red, 4=buzzer, 5=relay, 6=blue
        # mode 2=via RS autoreset, 3=via RS, 4=internal
        # times are in 1/10s
        self.bus.do_command(self.address, 0x50, [io, 0, 3, mode, time, time0, time1])


    def SetAutoconfig(self, enable, enableSend):
        time = 20 # 2sec
        format = 0x4
        buzz = 1        # 0=disable, 1=on first, 2=always
        self.bus.do_command(self.address, 0x58, [(0x1 if enable else 0), time, (0x1 if enableSend else 0), format, buzz])


    def SetRS485(self, newaddress, baudrate):
        if baudrate==2400:  b = 1
        elif baudrate==4800:  b = 2
        elif baudrate==9600:  b = 3
        elif baudrate==19200: b = 4 
        elif baudrate==38400: b = 5 
        elif baudrate==57600: b = 6 
        elif baudrate==115200: b = 7
        else: raise EValueError('Bad baudrate value')
        if (newaddress <= 0) or (newaddress > 254) :
            raise EValueError("BAd address value")

        self.bus.do_command(self.address, 0x54, [1, newaddress, b])

    
    def SetEventMask(self, enableCards, enableInput):
        self.bus.do_command(self.address, 0x7c, [(0x7 if enableCards else 0), 0,0,0, (0x1 if enableInput else 0)])


    def SetIO(self, io, value):
        self.bus.do_command(self.address, 0x70, [io, (1 if value else 0)])


    def GetEvent(self, pointer):
        frame = self.bus.do_command(self.address, 0x7a, [pointer & 0xff, (pointer >> 8) & 0xff])
        if not frame: return None

        return frame


    def GetEventStatus(self):
        frame = self.bus.do_command(self.address, 0x78, [])
        if not frame: return
        pointer = frame[5] + (frame[6] << 8)
        capacity = frame[3] + (frame[4] << 8)
        counter = frame[7] | (frame[8] << 8) | (frame[9] << 16) | (frame[10] << 24)
        if counter == 0xffffffff: 
            counter = 0 
        else:
            counter += 1

        if not(self.lastEvent is None):
            if counter > self.lastCounter:
                while self.lastCounter < counter: 
                    frame =  self.GetEvent(self.lastEvent)
                    if frame: 
                        if self.Worker:
                            self.Worker(self.address, ParseEvent(frame))
                        else:
                            PrintFrame(frame)
                    self.lastEvent = (self.lastEvent + 1) % capacity 
                    self.lastCounter += 1

        self.lastEvent = pointer
        self.lastCounter = counter



def setupReader(address):
    '''
        set NFC Reader with expected default settings
    '''
    bus = NetronixBus('/dev/ttyNS0', baudrate=9600)
    d = NetronixDevice(bus=bus, address=1)
    d.Login()
    d.SetAutoconfig(True, False)
    d.SetAutowrite(True)
    d.SetEventMask(True, True)
    d.SetIOConfig(2,2,15,0,15) # green
    d.SetIOConfig(3,2,12,3,1) # red
    d.SetIOConfig(4,2,1,0,1)  #buzzer
    d.SetIOConfig(5,0,1,0,1)  #relay
    d.SetIOConfig(6,2,15,0,15) # blue
    d.SetRS485(address,baudrate=19200)
    d = NetronixBus('/dev/ttyNS0',baudrate=19200)
    d = NetronixDevice(bus=bus, address=address)
    d.GetEventStatus()

